;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "provide.rkt" "runtime.rkt" setup/getinfo)

(define-procedure-group/provide base-procedures define-base-procedure)

(define/provide true-string (make-parameter "true"))
(define/provide false-string (make-parameter "false"))
(define/provide current-yield (make-parameter #f))
(define/provide local-key (make-parameter 'local))
(define/provide locals-key (make-parameter 'locals))

(define (local-procedure-maker locals)
  (lambda (num [default #f])
    (let ([num* (string->integer (num))])
      ((with-handlers
           ([exn:fail:contract? (lambda _
                                  (cond [default]
                                        [else (raise-error/params
                                               'value-out-of-range num*)]))])
         (list-ref locals num*))))))

(define (locals-procedure-maker locals)
  (case-lambda
    [() (number->string (length locals))]
    [(key)
     (let ([yield (current-yield)])
       (for ([local locals])
         (call-extended (λ () (call-yield yield)) key local))
       "")]))

(define/provide (thunk-maker->procedure thunk-maker)
  (lambda locals
    ((thunk-maker (lambda (sym arguments)
                    (apply call-procedure*
                           (cond [(procedure-ref/forgive sym)]
                                 [(eq? sym (local-key))
                                  (procedure/properties
                                   (local-procedure-maker locals)
                                   (current-position) sym)]
                                 [(eq? sym (locals-key))
                                  (procedure/properties
                                   (locals-procedure-maker locals)
                                   (current-position) sym)]
                                 [else (raise-error/params
                                        'unknown-procedure sym
                                        'referenced-in (current-position))])
                           arguments))))))

(define-syntax/provide call-yield
  (syntax-rules ()
    [(_ thunk)
     (with-current-position
      (parameterize ([current-yield thunk])
        (call-procedure* (procedure-ref/raise 'yield))))]))

(define/provide (boolean->string b)
  ((if b true-string false-string)))

(define/provide (string->boolean s)
  (cond [(equal? s (true-string)) #t]
        [(equal? s (false-string)) #f]
        [else (raise-error/params 'invalid-boolean s)]))

(define/provide (string->number* str)
  (let ([x (string->number str)])
    (if (number? x)
        x
        (raise-error/params 'invalid-number str))))

(define/provide (string->divisor str)
  (let ([num (string->number* str)])
    (if (zero? num)
        (raise-error/params 'division-by-zero)
        num)))

(define/provide (string->integer str)
  (let ([x (string->number str)])
    (if (integer? x)
        x
        (raise-error/params 'invalid-integer str))))

(define/provide (string->natural str)
  (let ([x (string->integer str)])
    (if (negative? x)
        (raise-error/params 'value-out-of-range str)
        x)))

(define/provide (pregexp* str)
  (with-handlers ([exn:fail:contract? (error/params-raiser
                                       'invalid-pregexp str)])
    (pregexp str)))

(define/provide (extended hsh keys/values #:call-values? [c-v? #t])
  (let loop ([hsh hsh]
             [keys/values keys/values])
    (cond [(null? keys/values) hsh]
          [(null? (rest keys/values))
           (raise-error/params 'wrong-number-of-arguments)]
          [else (let ([name (string->symbol ((first keys/values)))]
                      [values/keys (rest keys/values)])
                  (loop (hash-set hsh name
                                  (procedure/properties
                                   (if c-v?
                                       (let ([v ((first values/keys))])
                                         (λ () v))
                                       (first values/keys))
                                   (procedure-position (first keys/values))
                                   name))
                        (rest values/keys)))])))

(define/provide (call-extended* thunk keys/values #:call-values? [c-v? #t])
  (parameterize ([current-procedures
                  (cons (extended (first (current-procedures))
                                  keys/values #:call-values? c-v?)
                        (rest (current-procedures)))])
    (thunk)))

(define/provide (call-extended thunk . keys/values)
  (call-extended* thunk keys/values #:call-values? #f))

(define/provide (make-match-procedure match)
  (lambda ([num (const "0")])
    (let ([num* (string->integer (num))])
      (with-handlers ([exn:fail:contract?
                       (lambda _ (raise-error/params
                                  'value-out-of-range num*))])
        (cond [(list-ref match num*)]
              [else ""])))))

(define-base-procedure help
  "{help P} returns the documentation of the procedure P.
   Example: ${help procedures}"
  (case-lambda
    [() (procedure-ref/call 'help (λ () "help"))]
    [(key)
     (let ([proc (procedure-ref (string->symbol (key)))])
       (string-replace (procedure-doc proc) "\n   "  "\n"))]))

(define-base-procedure procedures
  "{procedures} returns the list of defined procedures separated by
   newlines.  {procedures P} iterates the list of defined procedures; for
   each procedure, temporarily defines a procedure named P that returns
   the name of the current procedure, and calls {yield}.  For example,
   ${procedures} returns the same as:
   ${procedures p | join {p} {char 10}}"
  (let ([procedures-list
         (lambda ()
           (sort (set-map (foldl (lambda (hsh s)
                                   (set-union (list->set (hash-keys hsh))
                                              s))
                                 (set) (current-procedures))
                          symbol->string)
                 string<?))])
    (case-lambda
      [() (string-join (procedures-list) "\n")]
      [(key)
       (let ([yield (current-yield)])
         (for ([name (procedures-list)])
           (call-extended (λ () (call-yield yield)) key (λ () name)))
         "")])))

(define-base-procedure source
  "{source P} returns the location of the definition of the procedure P."
  (lambda (key)
    (position->string (procedure-position
                       (procedure-ref (string->symbol (key)))))))

(define-base-procedure |#|
  "{# ...} ignores its arguments and returns an empty string.  It may
   be used to comment code out."
  (const ""))

(define-base-procedure and
  "{and} returns true.  {and A} returns true if A is true; if it's
   false, returns false.  {and A1 A2 ... An} is equivalent to {and A2
   ... An} if A1 is true; if it's false, returns false."
  (lambda thunks
    (boolean->string (andmap (λ (t) (string->boolean (t))) thunks))))

(define-base-procedure or
  "{or} returns false.  {or A} returns true if A is true; if it's
   false, returns false.  {or A1 A2 ... An} is equivalent to {or A2
   ... An} if A1 is false; if it's true, returns true."
  (lambda thunks
    (boolean->string (ormap (λ (t) (string->boolean (t))) thunks))))

(define-base-procedure defined?
  "{defined? P} returns true if the procedure P is defined, otherwise
   returns false."
  (lambda (thunk)
    (boolean->string (procedure-ref/forgive (string->symbol (thunk))))))

(define (comparator test (get (λ (x) (x))))
  (lambda thunks
    ((let loop ([thunks thunks])
       (cond [(null? thunks) true-string]
             [(null? (rest thunks)) true-string]
             [(test (get (first thunks)) (get (second thunks)))
              (loop (rest thunks))]
             [else false-string])))))

(define-base-procedure string=?
  "{string=?} and {string=? A} return true.  {string=? A1 A2 ... An}
   is equivalent to {string=? A2 ... An} if A1 and A2 are the same
   string; otherwise, returns false."
  (comparator string=?))

(define-base-procedure string<?
  "{string<?} and {string<? A} return true.  {string<? A1 A2 ... An}
   is equivalent to {string<? A2 ... An} if A1 precedes A2 in
   lexicographical order; otherwise, returns false."
  (comparator string<?))

(define-base-procedure string<=?
  "{string<=?} and {string<=? A} return true.  {string<=? A1 A2
   ... An} is equivalent to {string<=? A2 ... An} if A1 precedes or
   matches A2 in lexicographical order; otherwise, returns false."
  (comparator string<=?))

(define-base-procedure string>?
  "{string>?} and {string>? A} return true.  {string>? A1 A2 ... An}
   is equivalent to {string>? A2 ... An} if A1 goes after A2 in
   lexicographical order; otherwise, returns false."
  (comparator string>?))

(define-base-procedure string>=?
  "{string>=?} and {string>=? A} return true.  {string>=? A1 A2
   ... An} is equivalent to {string>=? A2 ... An} if A1 goes after or
   matches A2 in lexicographical order; otherwise, returns false."
  (comparator string>=?))

(define-base-procedure not
  "{not A} returns false if A is true; if it's false, returns true."
  (λ (thunk) (boolean->string (not (string->boolean (thunk))))))

(define-base-procedure if
  "{if B C A} returns C if B is true; if it's false, returns A.  {if
   B C} is equivalent to {if B C ''}."
  (lambda (test consequent [alternate (const "")])
    ((if (string->boolean (test)) consequent alternate))))

(define-base-procedure case
  "{case A K V} returns V if A is the same string as K; otherwise,
   signals an error.  {case A K V D} returns D instead of signaling an
   error.  {case A K1 V1 K2 V2 ...} returns V1 if A is the same string as
   K1; otherwise, is equivalent to {case A K2 V2 ...}."
  (lambda (str . keys/values)
    (let ([str* (str)])
      (let loop ([keys/values keys/values])
        (cond [(null? keys/values)
               (raise-error/params 'unknown-value str*)]
              [(null? (rest keys/values)) ((first keys/values))]
              [(equal? ((first keys/values)) str*)
               ((second keys/values))]
              [else (loop (rest (rest keys/values)))])))))

(define-base-procedure yield
  "{yield} calls the sequence element processing coroutine, which is
   set up temporarily by procedures like `join` or `for-each`.  Returns
   an empty string."
  (lambda ()
    (cond [(current-yield) => (lambda (yield)
                                (parameterize ([current-yield #f])
                                  (yield)))]
          [else (raise-error/params 'unexpected-sequence)])
    ""))

(define-base-procedure length
  "{length A} returns the length of the sequence produced by A, if
   it's greater than 0; otherwise, returns the length of the string
   returned by A."
  (lambda (str-or-sequence)
    (let* ([count 0]
           [str (parameterize ([current-yield
                                (λ () (set! count (add1 count)))])
                  (str-or-sequence))])
      (number->string (if (zero? count)
                          (string-length str)
                          count)))))

(define-base-procedure +
  "{+} returns 0.  {+ Z ...} returns the sum of its arguments."
  (lambda thunks
    (number->string (foldl (λ (t n) (+ n (string->number* (t))))
                           0 thunks))))

(define-base-procedure -
  "{- Z W} returns the subtraction of W from Z.  {- W} is equivalent
   to {- 0 W}.  {- Z W ...} is equivalent to {- Z {+ W ...}}."
  (lambda (thunk . thunks)
    (number->string (if (null? thunks)
                        (- (string->number* (thunk)))
                        (- (string->number* (thunk))
                           (foldl (λ (t n) (+ n (string->number* (t))))
                                  0 thunks))))))

(define-base-procedure *
  "{*} returns 1.  {* Z ...} returns the product of its arguments."
  (lambda thunks
    (number->string (foldl (λ (t n) (* n (string->number* (t))))
                           1 thunks))))

(define-base-procedure /
  "{/ Z W} returns the division of Z by W.  {/ W} is equivalent to
   {/ 1 W}.  {/ Z W1 W2 ... Wn} is equivalent to {/ {/ Z W1} W2 ... Wn}."
  (lambda (thunk . thunks)
    (number->string (if (null? thunks)
                        (/ (string->divisor (thunk)))
                        (foldl (λ (t n) (/ n (string->divisor (t))))
                               (string->number* (thunk)) thunks)))))

(define-base-procedure %
  "{% N M} returns N modulo M."
  (lambda (dividend divisor)
    (number->string (modulo (string->number* (dividend))
                            (string->divisor (divisor))))))

(define (number-converter conv-proc)
  (λ (proc) (number->string (conv-proc (string->number* (proc))))))

(define-base-procedure inexact
  "{inexact Z} returns the coercion of Z into an inexact number. "
  (number-converter exact->inexact))

(define-base-procedure exact
  "{exact Z} returns the coercion of Z into an exact number. "
  (number-converter inexact->exact))

(define-base-procedure truncate
  "{truncate X} returns the integer farthest from 0 that isn't
   farther from 0 than X."
  (number-converter truncate))

(define-base-procedure round
  "{round X} returns the integer closest to X, resolving ties in favor
   of an even number."
  (number-converter round))

(define-base-procedure floor
  "{floor X} returns the largest integer that is no more than X."
  (number-converter floor))

(define-base-procedure ceiling
  "{ceiling X} returns the smallest integer that is at least as large as X."
  (number-converter ceiling))

(define-base-procedure abs
  "{abs X} returns the absolute value of X."
  (number-converter abs))

(define (number-discriminator better?)
  (lambda (thunk . thunks)
    (number->string (foldl (lambda (t best)
                             (let ([n (string->number* (t))])
                               (if (better? n best) n best)))
                           (string->number* (thunk)) thunks))))

(define-base-procedure min
  "{min X ...} compares its arguments as numbers and returns the smallest."
  (number-discriminator <))

(define-base-procedure max
  "{max X ...} compares its arguments as numbers and returns the largest."
  (number-discriminator >))

(define (number-comparator test)
  (comparator test (λ (x) (string->number* (x)))))

(define-base-procedure =
  "{=} and {= X} return true.  {= X1 X2 ... Xn} is equivalent to {= X2
   ... Xn} if X1 is numerically equal to X2; otherwise, returns false."
  (number-comparator =))

(define-base-procedure >
  "{>} and {> X} return true.  {> X1 X2 ... Xn} is equivalent to {> X2
   ... Xn} if X1 is numerically greater than X2; otherwise, returns
   false."
  (number-comparator >))

(define-base-procedure <
  "{<} and {< X} return true.  {< X1 X2 ... Xn} is equivalent to {< X2
   ... Xn} if X1 is numerically less than X2; otherwise, returns false."
  (number-comparator <))

(define-base-procedure >=
  "{>=} and {>= X} return true.  {>= X1 X2 ... Xn} is equivalent to
   {>= X2 ... Xn} if X1 is numerically equal or greater than X2;
   otherwise, returns false."
  (number-comparator >=))

(define-base-procedure <=
  "{<=} and {<= X} return true.  {<= X1 X2 ... Xn} is equivalent to
   {<= X2 ... Xn} if X1 is numerically equal or less than X2; otherwise,
   returns false."
  (number-comparator <=))

(define-base-procedure random
  "{random} returns a random real number between 0 and 1.  {random A}
   returns a random integer between 0 (included) and A (excluded) if it
   isn't an empty string, otherwise returns an empty string.  It also
   randomizes the order of the sequence produced by A."
  (case-lambda
    [() (number->string (random))]
    [(num-or-sequence)
     (let* ([params '()]
            [str (parameterize ([current-yield
                                 (lambda ()
                                   (set! params
                                         (cons (current-parameterization)
                                               params)))])
                   (num-or-sequence))]
            [yield (current-yield)])
       (for ([param (shuffle params)])
         (call-with-parameterization param (λ () (call-yield yield))))
       (if (equal? "" str)
           ""
           (let ([num (string->integer str)])
             (number->string (with-handlers ([exn:fail:contract?
                                              (lambda _
                                                (raise-error/params
                                                 'value-out-of-range num))])
                               (random (inexact->exact num)))))))]))

(define-base-procedure for-each
  "{for-each S A} evaluates A for each element of the sequence S.  It's
   useful for conditionally calling {yield}, or calling procedures
   with side effects.  Returns an empty string."
  (lambda (sequence thunk)
    (let ([yield (current-yield)])
      (parameterize ([current-yield (lambda ()
                                      (parameterize ([current-yield yield])
                                        (thunk)))])
        (sequence))
      "")))

(define-base-procedure join
  "{join S A B} collects A for each element of the sequence S, and
   returns its concatenation separated by B.  {join S A} is equivalent to
   {join S A ''}."
  (lambda (sequence thunk [sep (const "")])
    (let ([first? #t])
      (call-with-output-string
       (lambda (port)
         (parameterize ([current-yield (lambda ()
                                         (if first?
                                             (set! first? #f)
                                             (display (sep) port))
                                         (display (thunk) port))])
           (sequence)))))))

(define-base-procedure apply
  "{apply S P T} calls the procedure named P with as many arguments as
   elements in the sequence S, each one being T for the corresponding
   element of the sequence.  {apply S P A ... T} prepends A ... to the
   arguments to P.  Example: ${range i 10 | apply + -2 -1 {i}}"
  (lambda (sequence key . args)
    (when (null? args)
      (raise-error/params 'wrong-number-of-arguments))
    (procedure-ref/apply
     (string->symbol (key))
     (foldr (lambda (arg lst)
              (define (arg-proc)
                (let ([params (current-parameterization)])
                  (λ () (call-with-parameterization params arg))))
              (cond [(null? lst)
                     (parameterize
                         ([current-yield
                           (λ () (set! lst (cons (arg-proc) lst)))])
                       (sequence))
                     (reverse lst)]
                    [else (cons arg lst)]))
            '() args))))

(define-base-procedure digest
  "{digest S P N F} iterates the sequence S; for each element,
   temporarily defines a procedure named P and evaluates N.  {P P1 V1
   ... Pn Vn} temporarily defines the procedures P1 ... Pn with values V1
   ... Vn, which will be available in the next evaluation of N or F.
   Finally, F is returned.  {digest S P N F P1 V1 ... Pn Vn} is
   equivalent to {extended {digest S P N F} P1 V1 ... Pn Vn}.  Example:
   ${range i 10 | digest k {k j {+ {i} {j}}} {j} j 0}"
  (lambda (sequence key next final . keys/values)
    (let ([state (extended (hasheq) keys/values)])
      (parameterize
          ([current-yield
            (lambda ()
              (parameterize ([current-procedures
                              (cons state (current-procedures))])
                (call-extended next
                               key (lambda args
                                     (set! state (extended state args))
                                     ""))))])
        (sequence))
      (parameterize ([current-procedures (cons state (current-procedures))])
        (final)))))

(define-base-procedure any?
  "{any? S A} returns true if A is true for any element in the
   sequence S; otherwise, returns false."
  (lambda (sequence thunk)
    (let ([found? #f])
      (parameterize ([current-yield (lambda ()
                                      (when (and (not found?)
                                                 (string->boolean (thunk)))
                                        (set! found? #t)))])
        (sequence))
      (boolean->string found?))))

(define-base-procedure every?
  "{every? S A} returns true if A is true for every element in the
   sequence S; otherwise, returns false."
  (lambda (sequence thunk)
    (let ([found? #f])
      (parameterize ([current-yield (lambda ()
                                      (unless (string->boolean (thunk))
                                        (set! found? #t)))])
        (sequence))
      (boolean->string (not found?)))))

(define-base-procedure reverse
  "{reverse A} reverses both the order of the sequence produced and
   the string returned by A."
  (lambda (thunk)
    (let* ([params '()]
           [str (parameterize
                    ([current-yield
                      (lambda ()
                        (set! params (cons (current-parameterization)
                                           params)))])
                  (thunk))])
      (unless (null? params)
        (let ([yield (current-yield)])
          (for ([param params])
            (call-with-parameterization param (λ () (call-yield yield))))
          ""))
      (with-output-to-string
        (lambda ()
          (for ([i (range (sub1 (string-length str)) -1 -1)])
            (write-char (string-ref str i))))))))

(define-base-procedure sort
  "{sort S A P} produces a sequence by sorting the sequence S;
   compares each element by A using the procedure named P, which must
   accept at least 2 arguments and return true or false.  {sort S A} is
   equivalent to {sort S A string<?}."
  (lambda (sequence value [less? (const "string<?")])
    (let* ([params '()]
           [yield (current-yield)]
           [less?-proc (procedure-ref (string->symbol (less?)))])
      (parameterize ([current-yield
                      (lambda ()
                        (set! params (cons (cons (value)
                                                 (current-parameterization))
                                           params)))])
        (sequence))
      (for ([entry (sort params (lambda (x y)
                                  (string->boolean
                                   (call-procedure less?-proc (λ () (car x))
                                                   (λ () (car y))))))])
        (call-with-parameterization (cdr entry) (λ () (call-yield yield))))
      "")))

(define-base-procedure with-index
  "{with-index S P} iterates the sequence S; for each element,
   temporarily defines a procedure named P, which returns the index of
   the element, and calls {yield}."
  (lambda (sequence key)
    (let ([i 0]
          [yield (current-yield)])
      (parameterize ([current-yield
                      (lambda ()
                        (call-extended (lambda ()
                                         (set! i (add1 i))
                                         (call-yield yield))
                                       key (let ([str (number->string i)])
                                             (λ () str))))])
        (sequence)))))

(define-base-procedure nth
  "{nth S N A} returns A for the N-th element of the sequence S."
  (lambda (sequence num str)
    (let ([num* (string->integer (num))]
          [count 0]
          [str* #f])
      (parameterize ([current-yield (lambda ()
                                      (when (eqv? count num*)
                                        (set! str* (str)))
                                      (set! count (add1 count)))])
        (sequence))
      (if str* str* (raise-error/params 'value-out-of-range num*)))))

(define-base-procedure drop
  "{drop S N} produces a sequence by discarding N elements from the
   beginning of the sequence S, and returns the string returned by S
   without the first N characters."
  (lambda (sequence num)
    (let* ([num* (string->natural (num))]
           [i 0]
           [yield (current-yield)]
           [str (parameterize ([current-yield
                                (lambda ()
                                  (cond [(= i num*) (call-yield yield)]
                                        [else (set! i (add1 i))]))])
                  (sequence))])
      (substring str (min num* (string-length str))))))

(define-base-procedure drop-while
  "{drop-while S B} produces a sequence by discarding the elements of
   the beginning of the sequence S while B is true."
  (lambda (sequence condition)
    (let ([found? #f]
          [yield (current-yield)])
      (parameterize ([current-yield
                      (lambda ()
                        (cond [found? (call-yield yield)]
                              [(string->boolean (condition)) ""]
                              [else (set! found? #t)
                                    (call-yield yield)]))])
        (sequence)))))

(define-base-procedure take
  "{take S N} produces a sequence by taking N elements from the
   beginning of the sequence S, and returns the first N characters of the
   string returned by S."
  (lambda (sequence num)
    (let* ([num* (string->natural (num))]
           [i 0]
           [yield (current-yield)]
           [str (parameterize ([current-yield (lambda ()
                                                (when (< i num*)
                                                  (set! i (add1 i))
                                                  (call-yield yield)))])
                  (sequence))])
      (substring str 0 (min num* (string-length str))))))

(define-base-procedure take-while
  "{take-while S B} produces a sequence by taking the elements from
   the beginning of the sequence S while B is true."
  (lambda (sequence condition)
    (let ([found? #f]
          [yield (current-yield)])
      (parameterize ([current-yield (lambda ()
                                      (cond [found?]
                                            [(string->boolean (condition))
                                             (call-yield yield)]
                                            [else (set! found? #t)]))])
        (sequence)))))

(define-base-procedure slice
  "{slice S P N} splits the sequence S into subsequences of N
   elements; for each subsequence, temporarily defines a procedure named
   P, which produces a sequence of the elements in the subsequence, and
   calls {yield}."
  (lambda (sequence key num)
    (let ([num* (let ([n (string->integer (num))])
                  (if (positive? n)
                      n
                      (raise-error/params 'number-must-be-positive n)))]
          [states '()]
          [count 0]
          [yield (current-yield)])
      (define (flush)
        (call-extended
         (λ () (call-yield yield))
         key (lambda ()
               (let ([yield* (current-yield)]
                     [stack (current-stack)])
                 (for ([state (reverse states)])
                   (call-with-parameterization
                    state (lambda ()
                            (parameterize ([current-stack stack])
                              (call-yield yield*)))))
                 ""))))
      (parameterize ([current-yield
                      (lambda ()
                        (set! states (cons (current-parameterization)
                                           states))
                        (set! count (add1 count))
                        (when (= count num*)
                          (flush)
                          (set! count 0)
                          (set! states '())))])
        (sequence))
      (unless (null? states) (flush))
      "")))

(define-base-procedure group-by
  "{group-by S P A} groups the elements of the sequence S by A; for
   each group, temporarily defines a procedure named P, which produces a
   sequence of the elements in the group, and calls {yield}."
  (lambda (sequence key val)
    (let ([yield (current-yield)]
          [vals '()]
          [params (make-hash)])
      (parameterize ([current-yield
                      (lambda ()
                        (let ([val* (val)])
                          (unless (member val* vals)
                            (set! vals (cons val* vals)))
                          (hash-update! params val*
                                        (curry cons
                                               (current-parameterization))
                                        '())))])
        (sequence))
      (for ([val* (reverse vals)])
        (call-extended
         (λ () (call-yield yield))
         key (lambda ()
               (let ([yield* (current-yield)]
                     [stack (current-stack)])
                 (for ([param (reverse (hash-ref params val*))])
                   (call-with-parameterization
                    param (lambda ()
                            (parameterize ([current-stack stack])
                              (call-yield yield*)))))
                 ""))))
      "")))

(define-base-procedure filter
  "{filter S B} produces a sequence with the elements of the sequence
   S in which B is true."
  (lambda (sequence condition)
    (let ([yield (current-yield)])
      (parameterize ([current-yield (lambda ()
                                      (when (string->boolean (condition))
                                        (call-yield yield)))])
        (sequence)))))

(define-base-procedure concat
  "{concat A ...} produces a sequence by concatenating the sequences
   produced by its arguments, and returns the concatenation of the
   strings returned by its arguments."
  (lambda sequences
    (call-with-output-string (lambda (port)
                               (for ([sequence sequences])
                                 (display (sequence) port))))))

(define-base-procedure range
  "{range P A Z S} counts from A to Z, increasing or decreasing by S;
   for each number, temporarily defines a procedure named P, which
   returns the current number, and calls {yield}.  {range P Z} is
   equivalent to {range P 0 Z 1}.  {range P A Z} is equivalent to {range
   P A Z 1}."
  (let ([helper
         (lambda (key start end step)
           (if (zero? step)
               (raise-error/params 'step-must-not-be-zero)
               (let ([end* (string->integer (end))]
                     [yield (current-yield)])
                 (for ([i (in-range start end* step)])
                   (call-extended (λ () (call-yield yield))
                                  key (λ () (number->string i))))
                 "")))])
    (case-lambda
      [(key end)
       (helper key 0 end 1)]
      [(key start end)
       (helper key (string->integer (start)) end 1)]
      [(key start end step)
       (helper key (string->integer (start)) end
               (string->integer (step)))])))

(define-base-procedure chars
  "{chars A P} iterates the characters in A; for each character,
   temporarily defines a procedure named P, which returns the character,
   and calls {yield}."
  (lambda (thunk key)
    (let ([yield (current-yield)])
      (for ([c (thunk)])
        (call-extended (λ () (call-yield yield)) key (λ () (string c))))
      "")))

(define-base-procedure split
  "{split A P B} splits A into substrings by B; for each substring,
   temporarily defines a procedure named P, which returns the
   substring, and calls {yield}.  {split A P} is equivalent to {split A P
   ' '}."
  (lambda (thunk key [sep (const " ")])
    (let ([yield (current-yield)])
      (for ([part (string-split (thunk) (sep) #:trim? #f)])
        (call-extended (λ () (call-yield yield)) key (λ () part)))
      "")))

(define-base-procedure pregexp-match?
  "{pregexp-match? A R} returns true if A matches the regular
   expression R; otherwise returns false."
  (λ (thunk exp) (boolean->string (regexp-match? (pregexp* (exp)) (thunk)))))

(define-base-procedure pregexp-match
  "{pregexp-match A P R} iterates the matches of the regular
   expression R in A; for each match, temporarily defines a procedure
   named P, and calls {yield}.  {P 0} returns the match substring.  {P}
   is equivalent to {P 0}.  {P N}, where N is greater than 0, returns the
   match group number N."
  (lambda (thunk key exp)
    (let ([yield (current-yield)])
      (for ([match (regexp-match* (pregexp* (exp)) (thunk)
                                  #:match-select identity)])
        (call-extended (λ () (call-yield yield))
                       key (make-match-procedure match)))
      "")))

(define-base-procedure pregexp-split
  "{pregexp-split A P R} splits A into substrings by the regular
   expression R; for each substring, temporarily defines a procedure
   named P, which returns the substring, and calls {yield}.
   {pregexp-split A P} is equivalent to {pregexp-split A P '\\\\s+'}."
  (lambda (thunk key [sep (const "\\s+")])
    (let ([yield (current-yield)])
      (for ([part (regexp-split (pregexp* (sep)) (thunk))])
        (call-extended (λ () (call-yield yield)) key (λ () part)))
      "")))

(define-base-procedure args
  "{args P A ...} for each A, temporarily defines a procedure named P,
   which returns A, and calls {yield}.  Example:
   ${args x 1 2 3 | join {x} ,}"
  (lambda (key . args)
    (let ([yield (current-yield)])
      (for ([arg args])
        (call-extended (λ () (call-yield yield)) key arg))
      "")))

(define-base-procedure extended
  "{extended A P1 V1 ... Pn Vn} temporarily defines the procedures
   named P1 ... Pn, which return the corresponding and previously
   evaluated V1 ... Vn, and returns A.  Example:
   ${extended {+ {a} {b}} a 1 b 2}"
  (lambda (thunk . keys/values)
    (call-extended* thunk keys/values)))

(define-base-procedure extended*
  "{extended* A P1 V1 ... Pn Vn} temporarily defines the procedures
   named P1 ... Pn, which return the corresponding V1 ... Vn, and returns
   A.  Example: ${extended* {+ {a} {b}} a 2 b {* {a} 2}}"
  (lambda (thunk . keys/values)
    (call-extended* thunk keys/values #:call-values? #f)))

(define-base-procedure ord
  "{ord A} returns the integer ordinal of the one-character string A."
  (lambda (thunk)
    (let ([str (thunk)])
      (if (= 1 (string-length str))
          (number->string (char->integer (string-ref str 0)))
          (raise-error/params 'string-must-be-of-length-1)))))

(define-base-procedure char
  "{char A} returns the character with ordinal A.
   {char A1 A2 ... An} is equivalent to {char A1}{char A2 ... An}.
   Example: ${char 72 69 76 76 79}"
  (lambda nums
    (apply string
           (map (lambda (num)
                  (let ([num* (string->integer (num))])
                    (with-handlers
                        ([exn:fail:contract? (lambda _
                                               (raise-error/params
                                                'value-out-of-range num*))])
                      (integer->char num*))))
                nums))))

(define-base-procedure replace*
  "{replace* A F T} returns A with the first occurrence of F replaced by T."
  (λ (str old new) (string-replace (str) (old) (new) #:all? #f)))

(define-base-procedure replace
  "{replace A F T} returns A with all the occurrences of F replaced by T."
  (λ (str old new) (string-replace (str) (old) (new) #:all? #t)))

(define (make-pregexp-replace proc)
  (case-lambda
    [(input exp insert)
     (proc (pregexp* (exp)) (input) (insert))]
    [(input key exp insert)
     (proc (pregexp* (exp)) (input)
           (lambda match
             (call-extended insert key (make-match-procedure match))))]))

(define-base-procedure pregexp-replace*
  "{pregexp-replace* A R I} returns A with the first occurrence of the
   regular expression R replaced by I.  {pregexp-replace* A P R I}
   temporarily defines a procedure named P before evaluating I.  {P 0}
   returns the match substring.  {P} is equivalent to {P 0}.  {P N},
   where N is greater than 0, returns the match group number N."
  (make-pregexp-replace regexp-replace))

(define-base-procedure pregexp-replace
  "{pregexp-replace A R I} returns A with all the occurrences of the
   regular expression R replaced by I.  {pregexp-replace A P R I}
   temporarily defines a procedure named P before evaluating I for each
   match.  {P 0} returns the match substring.  {P} is equivalent to {P
   0}.  {P N}, where N is greater than 0, returns the match group number
   N."
  (make-pregexp-replace regexp-replace*))

(define-base-procedure downcase
  "{downcase A} returns the downcase conversion of A."
  (λ (thunk) (string-downcase (thunk))))

(define-base-procedure upcase
  "{upcase A} returns the upcase conversion of A."
  (λ (thunk) (string-upcase (thunk))))

(define-base-procedure titlecase
  "{titlecase A} returns the titlecase conversion of A."
  (λ (thunk) (string-titlecase (thunk))))

(define-base-procedure normalize-nfkd
  "{normalize-nfkd A} returns the Unicode normalized form KD of A."
  (λ (thunk) (string-normalize-nfkd (thunk))))

(define-base-procedure normalize-nfc
  "{normalize-nfc A} returns the Unicode normalized form C of A."
  (λ (thunk) (string-normalize-nfc (thunk))))

(define-base-procedure normalize-nfkc
  "{normalize-nfkc A} returns the Unicode normalized form KC of A."
  (λ (thunk) (string-normalize-nfkc (thunk))))

(define-base-procedure normalize-nfd
  "{normalize-nfd A} returns the Unicode normalized form D of A."
  (λ (thunk) (string-normalize-nfd (thunk))))

(define-base-procedure version
  "{version} returns the Pinaret version."
  (λ () ((get-info '("pinaret")) 'version)))

(define/provide (date->iso8601-string date)
  (define (z x)
    (if (< x 10)
        (string-append "0" (number->string x))
        x))
  (format "~a-~a-~aT~a:~a:~a~a"
          (date-year date) (z (date-month date)) (z (date-day date))
          (z (date-hour date)) (z (date-minute date)) (z (date-second date))
          (let ([delta (date-time-zone-offset date)])
            (if (zero? delta)
                "Z"
                (let ([hours (quotient delta 3600)]
                      [minutes (modulo (quotient delta 60) 60)])
                  (string-append (if (negative? delta) "-" "+")
                                 (z (abs hours)) (z minutes)))))))

(define-base-procedure current-timestamp
  "{current-timestamp} returns the current timestamp in the ISO-8601
   year-month-day-hour-minute-second-timezone format.  The resulting
   string reflects the time in UTC, but {current-timestamp true} reflects
   the time according to the local time zone."
  (lambda ([local? false-string])
    (date->iso8601-string (seconds->date (current-seconds)
                                         (string->boolean (local?))))))
