;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "runtime.rkt" "procedure.rkt" "repl.rkt" "dict.rkt"
         "base.rkt" "system.rkt" rackunit rackunit/text-ui)
(require/expose
 "system.rkt"
 (cache-library
  cache-generators
  read-dict-directory))
(require/expose "procedure.rkt" (object->thunk-maker))

(define (string->thunk str)
  (thunk-maker->thunk (string->thunk-maker str)))

(define (string->procedure str)
  (thunk-maker->procedure (string->thunk-maker str)))

(define (T . x) ((string->thunk (apply string-append x))))

(define-check (check-e-p thunk sym)
  (unless (with-handlers ([error/params?
                           (lambda (e)
                             (andmap (curryr member (error/params-args e))
                                     (if (pair? sym) sym (list sym))))]
                          [exn? void])
            (thunk)
            #f)
    (fail-check)))

(define-check (check-t-s str)
  (unless (string->boolean str) (fail-check)))

(define-check (check-f-s str)
  (when (string->boolean str) (fail-check)))

(define-test-suite runtime-suite
  (current-procedures (list base-procedures))
  (test-case "recursion limit"
             (parameterize
                 ([current-procedures
                   (cons (hash-set (first (current-procedures))
                                   'r (λ () (procedure-ref/call 'r)))
                         (rest (current-procedures)))])
               (check-e-p (λ () (procedure-ref/call 'r))
                          'reached-recursion-limit))))

(define-test-suite base-suite
  (current-procedures (list base-procedures))
  (test-case "#"
             (check-equal? (T "${# foo bar}") "")
             (check-t-s (T "${string=? x x{# foo}}"))
             (check-f-s (T "${string=? x {# foo}}")))
  (test-case "and"
             (check-t-s (T "${and}"))
             (check-t-s (T "${and true}"))
             (check-f-s (T "${and false}"))
             (check-t-s (T "${and true true}"))
             (check-f-s (T "${and false true}"))
             (check-f-s (T "${and true false}"))
             (check-t-s (T "${and true true true}"))
             (check-f-s (T "${and false {xyz}}"))
             (check-e-p (λ () (T "${and x}")) 'invalid-boolean))
  (test-case "or"
             (check-f-s (T "${or}"))
             (check-t-s (T "${or true}"))
             (check-f-s (T "${or false}"))
             (check-t-s (T "${or true true}"))
             (check-t-s (T "${or false true}"))
             (check-f-s (T "${or false false}"))
             (check-t-s (T "${or false false true}"))
             (check-f-s (T "${or false false false}"))
             (check-t-s (T "${or true {xyz}}"))
             (check-e-p (λ () (T "${or x}")) 'invalid-boolean))
  (test-case "defined?"
             (check-t-s (T "${defined? defined?}"))
             (check-f-s (T "${defined? xyz}")))
  (test-case "string=?"
             (check-t-s (T "${string=?}"))
             (check-t-s (T "${string=? x}"))
             (check-t-s (T "${string=? x x}"))
             (check-f-s (T "${string=? x y}"))
             (check-t-s (T "${string=? x x x}"))
             (check-f-s (T "${string=? x x y}")))
  (test-case "string<?"
             (check-t-s (T "${string<? a b c}"))
             (check-f-s (T "${string<? a a b}"))
             (check-f-s (T "${string<? c b a}")))
  (test-case "string<=?"
             (check-t-s (T "${string<=? a b c}"))
             (check-t-s (T "${string<=? a a b}"))
             (check-f-s (T "${string<=? b a a}")))
  (test-case "string>?"
             (check-t-s (T "${string>? c b a}"))
             (check-f-s (T "${string>? b a a}"))
             (check-f-s (T "${string>? a b c}")))
  (test-case "string>=?"
             (check-t-s (T "${string>=? c b a}"))
             (check-t-s (T "${string>=? b a a}"))
             (check-f-s (T "${string>=? a b c}")))
  (test-case "equal?"
             (check-t-s (T "${not {string=? x y}}"))
             (check-f-s (T "${not {string=? x x}}"))
             (check-e-p (λ () (T "${not x}")) 'invalid-boolean))
  (test-case "if"
             (check-equal? (T "${if {string=? x x} x {xyz}}") "x")
             (check-equal? (T "${if {string=? x x} x}") "x")
             (check-equal? (T "${if {string=? x y} {xyz} x}") "x")
             (check-equal? (T "${if {string=? x y} {xyz}}") "")
             (check-e-p (λ () (T "${if x x}")) 'invalid-boolean))
  (test-case "case"
             (check-e-p (λ () (T "${case x}")) 'unknown-value)
             (check-equal? (T "${case x y}") "y")
             (check-e-p (λ () (T "${case x y z}")) 'unknown-value)
             (check-equal? (T "${case x y z a}") "a")
             (check-e-p (λ () (T "${case x y z a b}")) 'unknown-value)
             (check-equal? (T "${case x y z a b x c}") "c"))
  (test-case "length"
             (check-equal? (T "${length abc}") "3")
             (check-equal? (T "${length \"\"}") "0")
             (check-equal? (T "${length {range a 10}}") "10"))
  (test-case "+"
             (check-equal? (T "${+}") "0")
             (check-equal? (T "${+ 2}") "2")
             (check-equal? (T "${+ 2 1}") "3")
             (check-equal? (T "${+ 2 1 3}") "6")
             (check-e-p (λ () (T "${+ x}")) 'invalid-number))
  (test-case "-"
             (check-e-p (λ () (T "${-}")) 'wrong-number-of-arguments)
             (check-equal? (T "${- 2}") "-2")
             (check-equal? (T "${- 2 1}") "1")
             (check-equal? (T "${- 2 1 3}") "-2")
             (check-e-p (λ () (T "${- x}")) 'invalid-number))
  (test-case "*"
             (check-equal? (T "${*}") "1")
             (check-equal? (T "${* 3}") "3")
             (check-equal? (T "${* 3 2}") "6")
             (check-equal? (T "${* 3 2 4}") "24")
             (check-e-p (λ () (T "${* x}")) 'invalid-number))
  (test-case "/"
             (check-e-p (λ () (T "${/}")) 'wrong-number-of-arguments)
             (check-e-p (λ () (T "${/ 0}")) 'division-by-zero)
             (check-e-p (λ () (T "${/ 2 0}")) 'division-by-zero)
             (check-equal? (T "${/ 2}") "1/2")
             (check-equal? (T "${/ {/ 3 2} 4}") (T "${/ 3 2 4}"))
             (check-e-p (λ () (T "${/ x 1}")) 'invalid-number))
  (test-case "%"
             (check-equal? (T "${% 3 2}") "1")
             (check-e-p (λ () (T "${% x 1}")) 'invalid-number)
             (check-e-p (λ () (T "${% 2 0}")) 'division-by-zero))
  (test-case "inexact"
             (check-equal? (T "${inexact 1/2}") "0.5")
             (check-e-p (λ () (T "${inexact x}")) 'invalid-number))
  (test-case "exact"
             (check-equal? (T "${exact 0.5}") "1/2")
             (check-e-p (λ () (T "${exact x}")) 'invalid-number))
  (test-case "truncate"
             (check-equal? (T "${truncate 1.5}") "1.0")
             (check-equal? (T "${truncate 3/2}") "1")
             (check-e-p (λ () (T "${truncate x}")) 'invalid-number))
  (test-case "round"
             (check-equal? (T "${round 0.5}") "0.0")
             (check-equal? (T "${round 0.6}") "1.0")
             (check-equal? (T "${round 1.4}") "1.0")
             (check-equal? (T "${round 1.5}") "2.0")
             (check-equal? (T "${round 1/2}") "0")
             (check-equal? (T "${round 2/3}") "1")
             (check-e-p (λ () (T "${round x}")) 'invalid-number))
  (test-case "floor"
             (check-equal? (T "${floor 1.5}") "1.0")
             (check-equal? (T "${floor 3/2}") "1")
             (check-e-p (λ () (T "${floor x}")) 'invalid-number))
  (test-case "ceiling"
             (check-equal? (T "${ceiling 1.5}") "2.0")
             (check-equal? (T "${ceiling 3/2}") "2")
             (check-e-p (λ () (T "${ceiling x}")) 'invalid-number))
  (test-case "abs"
             (check-equal? (T "${abs -2}") "2")
             (check-equal? (T "${abs 2}") "2")
             (check-e-p (λ () (T "${abs x}")) 'invalid-number))
  (test-case "max"
             (check-e-p (λ () (T "${max}")) 'wrong-number-of-arguments)
             (check-equal? (T "${max 1}") "1")
             (check-equal? (T "${max 1 2}") "2")
             (check-equal? (T "${max 1 3 2}") "3")
             (check-equal? (T "${max 2 1 3}") "3")
             (check-e-p (λ () (T "${max x}")) 'invalid-number))
  (test-case "min"
             (check-e-p (λ () (T "${min}")) 'wrong-number-of-arguments)
             (check-equal? (T "${min 1}") "1")
             (check-equal? (T "${min 2 1}") "1")
             (check-equal? (T "${min 3 1 2}") "1")
             (check-equal? (T "${min 2 3 1}") "1")
             (check-e-p (λ () (T "${min x}")) 'invalid-number))
  (test-case "="
             (check-t-s (T "${=}"))
             (check-t-s (T "${= 1}"))
             (check-t-s (T "${= 1 1.0}"))
             (check-f-s (T "${= 1 2}"))
             (check-f-s (T "${= 1 1.0 2}"))
             (check-t-s (T "${= 1 1.0 1}"))
             (check-t-s (T "${= x}"))
             (check-e-p (λ () (T "${= 1 x}")) 'invalid-number))
  (test-case ">"
             (check-t-s (T "${>}"))
             (check-t-s (T "${> 1}"))
             (check-t-s (T "${> 2 1.0}"))
             (check-f-s (T "${> 1 2.0}"))
             (check-f-s (T "${> 2 1 1}"))
             (check-t-s (T "${> 3 2 1}"))
             (check-t-s (T "${> x}"))
             (check-e-p (λ () (T "${> 1 x}")) 'invalid-number))
  (test-case "<"
             (check-t-s (T "${<}"))
             (check-t-s (T "${< 1}"))
             (check-t-s (T "${< 1 2.0}"))
             (check-f-s (T "${< 2 1.0}"))
             (check-f-s (T "${< 1 2 2}"))
             (check-t-s (T "${< 1 2 3}"))
             (check-t-s (T "${< x}"))
             (check-e-p (λ () (T "${< 1 x}")) 'invalid-number))
  (test-case ">="
             (check-t-s (T "${>=}"))
             (check-t-s (T "${>= 1}"))
             (check-t-s (T "${>= 2 1.0}"))
             (check-t-s (T "${>= 2 2.0}"))
             (check-f-s (T "${>= 1 2.0}"))
             (check-t-s (T "${>= 2 1 1}"))
             (check-t-s (T "${>= 3 2 1}"))
             (check-t-s (T "${>= x}"))
             (check-e-p (λ () (T "${>= 1 x}")) 'invalid-number))
  (test-case "<="
             (check-t-s (T "${<=}"))
             (check-t-s (T "${<= 1}"))
             (check-t-s (T "${<= 1 2.0}"))
             (check-t-s (T "${<= 2 2.0}"))
             (check-f-s (T "${<= 2 1.0}"))
             (check-t-s (T "${<= 1 2 2}"))
             (check-t-s (T "${<= 1 2 3}"))
             (check-t-s (T "${<= x}"))
             (check-e-p (λ () (T "${<= 1 x}")) 'invalid-number))
  (test-case "random"
             (check-equal? (T "${random 1}") "0")
             (let ([x (T "${random 2}")])
               (check-true (or (equal? "0" x) (equal? "1" x))))
             (check-true (inexact? (string->number (T "${random}"))))
             (check-equal? (T "${random 1.0}") "0")
             (check-e-p (λ () (T "${random 1.1}")) 'invalid-integer)
             (check-e-p (λ () (T "${random -1}")) 'value-out-of-range)
             (check-e-p (λ () (T "${random 0}")) 'value-out-of-range)
             (check-true (<= 0 (string->number (T "${random 4294967087}"))))
             (check-e-p (λ () (T "${random 4294967088}"))
                        'value-out-of-range)
             (let ([x (T "${range a 3 | random | join {a} ,}")])
               (check-true (or (equal? x "0,1,2")
                               (equal? x "0,2,1")
                               (equal? x "1,0,2")
                               (equal? x "1,2,0")
                               (equal? x "2,0,1")
                               (equal? x "2,1,0")))))
  (test-case "for-each"
             (check-equal? (T "${range a 4"
                              " | for-each {if {= {% {a} 2} 0} {yield}}"
                              " | join {a} , }")
                           "0,2")
             (check-e-p (λ () (T "${yield | for-each {yield}}"))
                        'unexpected-sequence))
  (test-case "join"
             (check-equal? (T "${range a 4 | join {a} ,}") "0,1,2,3")
             (check-equal? (T "${range a 4 | join {a}}") "0123")
             (check-equal? (T "${range a -1 | join {a} ,}") "")
             (check-e-p (λ () (T "${yield | join {yield}}"))
                        'unexpected-sequence))
  (test-case "apply"
             (check-e-p (λ () (T "${range a 6 | apply -}"))
                        'wrong-number-of-arguments)
             (check-equal? (T "${range a 6 | apply - {a}}") "-15")
             (check-equal? (T "${range a 6 | apply - 1 {a}}") "-14")
             (check-e-p (λ () (T "${yield | apply - {yield}}"))
                        'unexpected-sequence))
  (test-case "digest"
             (check-equal? (T "${range a 4"
                              " | digest k {k b {a}({b}) n {+ 1 {n}}}"
                              "   {b}{n}"
                              "   b * n 0}")
                           "3(2(1(0(*))))4")
             (check-e-p (λ () (T "${range a 4 | digest k {k b} {b}}"))
                        'wrong-number-of-arguments)
             (check-e-p (λ () (T "${yield | digest k {yield} _}"))
                        'unexpected-sequence))
  (test-case "any?"
             (check-t-s (T "${range a 4 | any? {string=? {a} 2}}"))
             (check-f-s (T "${range a -1 | any? {string=? {a} 2}}"))
             (check-f-s (T "${range a 4 | any? {string=? {a} 4}}"))
             (check-e-p (λ () (T "${yield | any? {yield}}"))
                        'unexpected-sequence))
  (test-case "every?"
             (check-f-s (T "${range a 4 | every? {string=? {a} 2}}"))
             (check-t-s (T "${range a -1 | every? {string=? {a} 2}}"))
             (check-t-s (T "${range a 2 3 | every? {string=? {a} 2}}"))
             (check-e-p (λ () (T "${yield | every? {yield}}"))
                        'unexpected-sequence))
  (test-case "reverse"
             (check-e-p (λ () (T "${yield | reverse}"))
                        'unexpected-sequence)
             (check-equal? (T "${range a 4 | reverse | join {a} ,}")
                           "3,2,1,0")
             (check-equal? (T "${range a -1 | reverse | join {a} ,}") "")
             (check-equal? (T "${reverse 123}") "321"))
  (test-case "sort"
             (check-equal? (T "${range a 4 | sort {a} > | join {a} ,}")
                           "3,2,1,0")
             (check-e-p (λ () (T "${range a 4 | sort {a} >}"))
                        'unexpected-sequence)
             (check-e-p (λ () (T "${yield | sort {yield}}"))
                        'unexpected-sequence))
  (test-case "with-index"
             (check-equal? (T "${range b 1 5 | with-index a"
                              " | join {a}{b} ,}")
                           "01,12,23,34")
             (check-e-p (λ () (T "${yield | with-index a}"))
                        'unexpected-sequence))
  (test-case "nth"
             (check-e-p (λ () (T "${range x 0 | nth 1 {x}}"))
                        'value-out-of-range)
             (check-equal? (T "${range x 10 | reverse | nth 1 {x}}") "8")
             (check-e-p (λ () (T "${yield | nth 0 {yield}}"))
                        'unexpected-sequence))
  (test-case "drop"
             (check-equal? (T "${drop foo 1}") "oo")
             (check-equal? (T "${drop 23 1}") "3")
             (check-e-p (λ () (T "${range a 2 | drop 1}"))
                        'unexpected-sequence)
             (check-equal? (T "${range a 4 | drop 2 | join {a} ,}") "2,3")
             (check-e-p (λ () (T "${drop y xyz}")) 'invalid-integer)
             (check-e-p (λ () (T "${drop 0 xyz}")) 'invalid-integer)
             (check-e-p (λ () (T "${drop xyz -1}")) 'value-out-of-range))
  (test-case "take"
             (check-equal? (T "${take foo 2}") "fo")
             (check-equal? (T "${take 23 1}") "2")
             (check-e-p (λ () (T "${yield | take 1}"))
                        'unexpected-sequence)
             (check-equal? (T "${range a 4 | take 2 | join {a} ,}") "0,1")
             (check-e-p (λ () (T "${take y xyz}")) 'invalid-integer)
             (check-e-p (λ () (T "${take 0 xyz}")) 'invalid-integer)
             (check-e-p (λ () (T "${take 0 -1}")) 'value-out-of-range))
  (test-case
   "drop-while"
   (check-equal? (T "${range a 4 | drop-while {< {a} 2} | join {a} ,}")
                 "2,3")
   (check-e-p (λ () (T "${yield | drop-while false}"))
              'unexpected-sequence)
   (check-e-p (λ () (T "${range a 4 | drop-while y | join {a} ,}"))
              'invalid-boolean)
   (check-e-p (λ () (T "${yield | drop-while {yield}}"))
              'unexpected-sequence))
  (test-case
   "take-while"
   (check-equal? (T "${range a 4 | take-while {< {a} 2} | join {a}}") "01")
   (check-e-p (λ () (T "${yield | take-while true}")) 'unexpected-sequence)
   (check-e-p (λ () (T "${range a 4 | take-while y | join {a}}"))
              'invalid-boolean)
   (check-e-p (λ () (T "${yield | take-while {yield}}"))
              'unexpected-sequence))
  (test-case "range"
             (check-e-p (λ () (T "${range a 1}")) 'unexpected-sequence)
             (check-equal? (T "${range a 1 | join {a} ,}") "0")
             (check-equal? (T "${range a 2 | join {a} ,}") "0,1")
             (check-equal? (T "${range a 1 2 | join {a} ,}") "1")
             (check-equal? (T "${range a 1 3 | join {a} ,}") "1,2")
             (check-equal? (T "${range a 1 6 2 | join {a} ,}") "1,3,5")
             (check-equal? (T "${range a -1 | join {a} ,}") "")
             (check-equal? (T "${range a -1 -4 -1 | join {a} ,}") "-1,-2,-3")
             (check-e-p (λ () (T "${range a x}")) 'invalid-integer)
             (check-e-p (λ () (T "${range a 0 x}")) 'invalid-integer)
             (check-e-p (λ () (T "${range a 0 0 x}")) 'invalid-integer)
             (check-e-p (λ () (T "${range a 0 0 0}"))
                        'step-must-not-be-zero))
  (test-case
   "slice"
   (check-e-p (λ () (T "${slice _ _ _}")) 'invalid-integer)
   (check-e-p (λ () (T "${slice _ _ 0}")) 'number-must-be-positive)
   (check-e-p (λ () (T "${slice _ _ -1}")) 'number-must-be-positive)
   (check-e-p (λ () (T "${range a 1 | slice _ 1}")) 'unexpected-sequence)
   (check-equal? (T "${range a 11 | slice b 3 | join {b | join {a} .} ,}")
                 "0.1.2,3.4.5,6.7.8,9.10")
   (check-equal? (T "${range a 11 | slice b 1 | join {b | join {a} .} ,}")
                 "0,1,2,3,4,5,6,7,8,9,10"))
  (test-case "group-by"
             (check-equal? (T "${range a 11 | group-by g {< 5 {a}}"
                              " | join {g | join {a} ,} ;}")
                           "0,1,2,3,4,5;6,7,8,9,10")
             (check-e-p (λ () (T "${yield | group-by g {yield}}"))
                        'unexpected-sequence))
  (test-case "filter"
             (check-e-p (λ () (T "${range a 6 | filter {= 0 {% {a} 2}}}"))
                        'unexpected-sequence)
             (check-equal? (T "${range a 6 | filter {= 0 {% {a} 2}}"
                              " | join {a} ,}")
                           "0,2,4")
             (check-e-p (λ () (T "${yield | filter {yield}}"))
                        'unexpected-sequence))
  (test-case "concat"
             (check-equal? (T "${concat}") "")
             (check-equal? (T "${concat foo bar}") "foobar")
             (check-equal? (T "${concat | join {x} ,}") "")
             (check-equal? (T "${range x 1 3 | concat {range x 3 5}"
                              " | join {x} ,}")
                           "1,2,3,4"))
  (test-case "chars"
             (check-e-p (λ () (T "${chars 123 a}")) 'unexpected-sequence)
             (check-equal? (T "${chars 123 a | join {a} ,}") "1,2,3")
             (check-equal? (T "${chars \"\" a | join {a} ,}") ""))
  (test-case
   "split"
   (check-e-p (λ () (T "${split 123 a}")) 'unexpected-sequence)
   (check-equal? (T "${split 123 a '' | join {a} ,}") ",1,2,3,")
   (check-equal? (T "${split '1 2 3' a | join {a} ,}") "1,2,3")
   (check-equal? (T "${split 1,2,3 a , | join {a} ' '}") "1 2 3")
   (check-equal? (T "${split ,1,2,3, a , | join {a} ' '}") " 1 2 3 ")
   (check-equal? (T "${split '  1 2 3  ' a | join {a} ,}") ",,1,2,3,,"))
  (test-case "pregexp-match?"
             (check-t-s (T "${pregexp-match? foobar ^foo}"))
             (check-f-s (T "${pregexp-match? foobar ^xfoo}"))
             (check-t-s (T "${pregexp-match? foobar bar$}"))
             (check-f-s (T "${pregexp-match? foobar xbar$}"))
             (check-t-s (T "${pregexp-match? foobar bar}"))
             (check-t-s (T "${pregexp-match? foobar foo}"))
             (check-t-s (T "${pregexp-match? foobar ooba}"))
             (check-t-s (T "${pregexp-match? ' ' \\\\s}"))
             (check-f-s (T "${pregexp-match? foobar xbar}"))
             (check-e-p (λ () (T "${pregexp-match? '' (}"))
                        'invalid-pregexp))
  (test-case
   "pregexp-match"
   (check-e-p (λ () (T "${pregexp-match 123 a .}")) 'unexpected-sequence)
   (check-e-p (λ () (T "${pregexp-match 123 a ( | join {a} ,}"))
              'invalid-pregexp)
   (check-e-p (λ () (T "${pregexp-match 1 a . | join {a b} ,}"))
              'invalid-integer)
   (check-e-p (λ () (T "${pregexp-match 1 a . | join {a -1} ,}"))
              'value-out-of-range)
   (check-e-p (λ () (T "${pregexp-match 1 a . | join {a 1} ,}"))
              'value-out-of-range)
   (check-equal? (T "${pregexp-match 1a2b3 a [a-z]([0-9]) | join {a 1} ,}")
                 "2,3")
   (let ([x (T "${pregexp-match 1a2b3 a [a-z]([0-9]) | join {a} ,}")])
     (check-equal? x "a2,b3")
     (check-equal? (T "${pregexp-match 1a2b3 a [a-z]([0-9]) | join {a 0} ,}")
                   x)))
  (test-case
   "pregexp-split"
   (check-e-p (λ () (T "${pregexp-split 123 a}")) 'unexpected-sequence)
   (check-equal? (T "${pregexp-split 123 a '' | join {a} ,}") ",1,2,3,")
   (check-equal? (T "${pregexp-split '1 2 3' a | join {a} ,}") "1,2,3")
   (check-equal? (T "${pregexp-split 1,2,3 a , | join {a} ' '}") "1 2 3")
   (check-equal? (T "${pregexp-split ,1,,2,3, a ,+ | join {a} ' '}")
                 " 1 2 3 ")
   (check-equal? (T "${pregexp-split '  1 2 3  ' a | join {a} ,}") ",1,2,3,")
   (check-e-p (λ () (T "${pregexp-split '' a ( | join {a} ,}"))
              'invalid-pregexp))
  (test-case "args"
             (check-e-p (λ () (T "${args a 1 2 3}")) 'unexpected-sequence)
             (check-equal? (T "${args a 1 2 3 | join {a} ,}") "1,2,3")
             (check-equal? (T "${args a | join {a} ,}") ""))
  (test-case "procedures"
             (parameterize
                 ([current-procedures
                   (list (hasheq 'z #f
                                 'yield (procedure-ref 'yield)
                                 'procedures (procedure-ref 'procedures)
                                 'a #f
                                 'join (procedure-ref 'join)))])
               (check-equal? (T "${procedures}")
                             "a\njoin\nprocedures\nyield\nz")
               (check-equal? (T "${procedures x | join {x} ,}")
                             "a,join,procedures,yield,z")))
  (test-case
   "extended"
   (check-equal? (T "${extended 1}") "1")
   (check-equal? (T "${a | extended a 1}") "1")
   (check-equal? (T "${a | extended a 1 a 2}") "2")
   (check-e-p (λ () (T "${a | extended a}")) 'wrong-number-of-arguments)
   (check-e-p (λ () (T "${a | extended b 1 a {b}}")) 'unknown-procedure)
   (check-e-p (λ () (T "${a | extended a {b} b 1}")) 'unknown-procedure)
   (check-e-p (λ () (T "${a | extended a 1 b}")) 'wrong-number-of-arguments)
   (check-e-p (λ () (T "${a | extended b 1 | extended a {b}}"))
              'unknown-procedure)
   (check-equal? (T "${a | extended a {b} | extended b 1}") "1"))
  (test-case
   "extended*"
   (check-equal? (T "${extended* 1}") "1")
   (check-equal? (T "${a | extended* a 1}") "1")
   (check-equal? (T "${a | extended* a 1 a 2}") "2")
   (check-e-p (λ () (T "${a | extended* a}")) 'wrong-number-of-arguments)
   (check-equal? (T "${a | extended* a {b} b 1}") "1")
   (check-e-p (λ () (T "${a | extended* a 1 b}")) 'wrong-number-of-arguments)
   (check-equal? (T "${a | extended* b 1 | extended* a {b}}") "1"))
  (test-case "ord"
             (check-equal? (T "${ord a}") "97")
             (check-e-p (λ () (T "${ord ''}")) 'string-must-be-of-length-1)
             (check-e-p (λ () (T "${ord aa}")) 'string-must-be-of-length-1))
  (test-case "char"
             (check-equal? (T "${char}") "")
             (check-equal? (T "${char 0}") "\x00")
             (check-equal? (T "${char 0 1}") "\x00\x01")
             (check-equal? (T "${char #xffff}") "\uffff")
             (check-e-p (λ () (T "${char a}")) 'invalid-integer)
             (check-e-p (λ () (T "${char -1}")) 'value-out-of-range)
             (check-e-p (λ () (T "${char 1000000000}")) 'value-out-of-range))
  (test-case
   "replace*"
   (check-equal? (T "${replace 012345012345 234 ABC}") "01ABC501ABC5")
   (check-equal? (T "${replace* 012345012345 234 ABC}") "01ABC5012345")
   (check-equal? (T "${replace* 012345012345 321 ABC}") "012345012345"))
  (test-case
   "pregexp-replace"
   (check-equal? (T "${pregexp-replace '1  2   3' \\\\s+ ,}") "1,2,3")
   (check-e-p (λ () (T "${pregexp-replace '' ( ''}")) 'invalid-pregexp)
   (check-e-p (λ () (T "${pregexp-replace abc m a {m a a}}"))
              'wrong-number-of-arguments)
   (check-e-p (λ () (T "${pregexp-replace abc m a {m a}}"))
              'invalid-integer)
   (check-e-p (λ () (T "${pregexp-replace abc m a {m 1}}"))
              'value-out-of-range)
   (check-e-p (λ () (T "${pregexp-replace abc m a {m -1}}"))
              'value-out-of-range)
   (check-e-p (λ () (T "${pregexp-replace . m ( _}"))
              'invalid-pregexp)
   (check-equal? (T "${pregexp-replace abcbabc m a(b)c {m | upcase}}")
                 "ABCbABC")
   (check-equal? (T "${pregexp-replace abcbabc m a(b)c {m 0 | upcase}}")
                 "ABCbABC")
   (check-equal? (T "${pregexp-replace abcbabc m a(b)c {m 1 | upcase}}")
                 "BbB")
   (check-equal?
    (T "${pregexp-replace abcbabc m (a)(B)?(b) {m 3}{m 2}{m 1}}")
    "bacbbac"))
  (test-case
   "pregexp-replace*"
   (check-equal? (T "${pregexp-replace* '1  2   3' \\\\s+ ,}") "1,2   3")
   (check-e-p (λ () (T "${pregexp-replace* '' ( ''}")) 'invalid-pregexp)
   (check-e-p (λ () (T "${pregexp-replace abc m a {m a a}}"))
              'wrong-number-of-arguments)
   (check-e-p (λ () (T "${pregexp-replace* abc m a {m a}}"))
              'invalid-integer)
   (check-e-p (λ () (T "${pregexp-replace* abc m a {m 1}}"))
              'value-out-of-range)
   (check-e-p (λ () (T "${pregexp-replace* abc m a {m -1}}"))
              'value-out-of-range)
   (check-e-p (λ () (T "${pregexp-replace* . m ( a}")) 'invalid-pregexp)
   (check-equal? (T "${pregexp-replace* abcbabc m a(b)c {m | upcase}}")
                 "ABCbabc")
   (check-equal? (T "${pregexp-replace* abcbabc m a(b)c {m 0 | upcase}}")
                 "ABCbabc")
   (check-equal? (T "${pregexp-replace* abcbabc m a(b)c {m 1 | upcase}}")
                 "Bbabc")
   (check-equal?
    (T "${pregexp-replace* abcbabc m (a)(B)?(b) {m 3}{m 2}{m 1}}")
    "bacbabc"))
  (test-equal? "dowcase" (T "${downcase FOO}") "foo")
  (test-equal? "upcase" (T "${upcase foo}") "FOO")
  (test-equal? "titlecase" (T "${titlecase foo}") "Foo")
  (test-equal? "normalize-nfkd" (T "${normalize-nfkd ẛ̣}") "ṩ")
  (test-equal? "normalize-nfc" (T "${normalize-nfc ẛ̣}") "ẛ̣")
  (test-equal? "normalize-nfkc" (T "${normalize-nfkc ẛ̣}") "ṩ")
  (test-equal? "normalize-nfd" (T "${normalize-nfd ẛ̣}") "ẛ̣")
  (test-case "current-timestamp"
             (check-true (>= (string-length (T "${current-timestamp true}"))
                             (string-length (T "${current-timestamp}"))
                             20))))

(define-procedure-group test-procedures define-test-procedure)
(define-test-procedure foo (string->thunk "bar"))
(define-test-procedure x (string->procedure "${y}"))
(define-test-procedure y (string->thunk "moo"))
(define-test-procedure z (string->thunk "${z1}"))
(define-test-procedure z1 (string->thunk "${z2}"))
(define-test-procedure z2 (string->thunk "x"))
(define-test-procedure q (string->thunk "f"))
(define-test-procedure f1 (λ (x) (string-upcase (x))))
(define-test-procedure f2 (λ (x) (string-titlecase (x))))
(define-test-procedure f3
  (λ xs (apply string-append (map (λ (x) (x)) xs))))
(define-test-procedure locals-test
  (string->procedure
   "${locals v | join {v} ,}-${local 0}-${local 1 _}-${locals}"))
(define-test-procedure recursion-test
  (string->procedure
   "${if {local 0} {recursion-test {not {local 0}}}}"))

(define procedure-string-sample
  (string-append "a${i{q} {and { de\"fine\"d? z  }\n"
                 "             {not {string=? \"{{z}}\" foo}}}\n"
                 "        \"|'{{z} |f1| f2 | f3 \"a\"}bar`\"\n"
                 "        foo\n"
                 "   |'f1'| f`2`}z"))

(define (m-p/v p l c v)
  (position/value (position p l c) v))

(define parsed-procedure-sample
  (list
   "a"
   (application
    (m-p/v #f 5 10 "f2")
    (list
     (m-p/v
      #f 5 4
      (application
       (m-p/v #f 5 4 "f1")
       (list
        (m-p/v
         #f 1 6
         (application
          (m-p/v #f 1 6 (list "i" (application (m-p/v #f 1 8 "q") '())))
          (list
           (m-p/v
            #f 1 11
            (application
             (m-p/v #f 1 12 "and")
             (list (m-p/v #f 1 16
                          (application (m-p/v #f 1 18 "defined?")
                                       (list (m-p/v #f 1 29 "z"))))
                   (m-p/v #f 2 13
                          (application
                           (m-p/v #f 2 14 "not")
                           (list
                            (m-p/v #f 2 18
                                   (application
                                    (m-p/v #f 2 19 "string=?")
                                    (list (m-p/v #f 2 28
                                                 (application
                                                  (m-p/v #f 2 30
                                                         (application
                                                          (m-p/v #f 2 31 "z")
                                                          '()))
                                                  '()))
                                          (m-p/v #f 2 36 "foo"))))))))))
           (m-p/v
            #f 3 8
            (list "|'"
                  (application
                   (m-p/v #f 3 26 "f3")
                   (list
                    (m-p/v #f 3 20
                           (application
                            (m-p/v #f 3 21 "f2")
                            (list (m-p/v #f 3 17
                                         (application
                                          (m-p/v #f 3 17 "f1")
                                          (list (m-p/v
                                                 #f 3 12
                                                 (application
                                                  (m-p/v #f 3 12
                                                         (application
                                                          (m-p/v #f 3 13 "z")
                                                          '()))
                                                  '()))))))))
                    (m-p/v #f 3 29 "a")))
                  "bar`"))
           (m-p/v #f 4 8 "foo")))))))))
   "z"))

(define-test-suite procedure-suite
  (current-procedures (list test-procedures base-procedures))
  (test-case
   "parse-procedure"
   (check-equal? (parse-procedure "x") "x")
   (check-equal? (parse-procedure procedure-string-sample #f 1 3)
                 parsed-procedure-sample)
   (check-equal? (T procedure-string-sample) "a|'Mooabar`z")
   (check-equal? (parse-procedure "$\\{") "${")
   (check-equal? (parse-procedure "$\\\\{") "$\\{")
   (check-equal? (parse-procedure "$\\\\\\{") "$\\\\{")
   (check-e-p (λ () (parse-procedure "${\"x}\"}"))
              'unexpected-end-of-procedure-application)
   (check-e-p (λ () (parse-procedure "${")) 'unfinished-procedure-application)
   (check-e-p (λ () (parse-procedure "${\"")) 'unfinished-quoting)
   (check-e-p (λ () (parse-procedure "${\\")) 'unfinished-escape-sequence)
   (check-e-p (λ () (parse-procedure "${}")) 'application-without-procedure)
   (check-e-p (λ () (parse-procedure "${a|}")) 'application-without-procedure)
   (check-e-p (λ () (parse-procedure "$\\")) 'unfinished-escape-sequence))
  (test-case "object->thunk-maker"
             (check-e-p (λ () (object->thunk-maker 1)) 'wrong-type))
  (test-case
   "locals"
   (check-equal? (T "${x y}") (T "${x}"))
   (check-equal? (T "${locals-test a b c}") "a,b,c-a-b-3")
   (check-equal? (T "${locals-test a}") "a-a-_-1")
   (check-e-p (λ () (T "${locals-test}")) 'value-out-of-range)
   (check-equal? (T "${recursion-test true}") (T "${recursion-test false}"))))

(define-test-suite dict-suite
  (test-case "parse-dict"
             (check-equal? (call-with-input-string
                            (string-append "\uFEFF  | || foo  \rbar \r\n "
                                           "| \\a\n\n|#\n|#\r\n|#\r|#\n\r\n"
                                           "|#\n\n|#\n\n\n|#\r\r\r|#\n\r\n\n"
                                           "| \t|\n | baz\rfoo bar\n\\|x"
                                           "\n\\\\|\n")
                            (curry parse-dict list* '()))
                           '("baz" 20 3 "foo bar\n|x\n\\|" 21 0
                             "#" 17 1 "\r\n" 17 0
                             "#" 14 1 "\r" 15 0
                             "#" 11 1 "\n" 12 0
                             "#" 9 1 "" 10 0
                             "#" 8 1 "" 8 0
                             "#" 7 1 "" 7 2
                             "#" 6 1 "" 7 0
                             "#" 5 1 "" 5 2
                             "foo" 1 7 " \rbar \r\n | \\a\n" 1 11)))
  (test-case "read-dict-file"
             (let ([x (read-dict-file "library/a.dict")])
               (check-equal? (hash-count x) 3)
               (check-equal? ((hash-ref x 'a)) "1")
               (check-equal? ((hash-ref x 'b)) "2")
               (check-equal? ((hash-ref x 'c)) "3")
               (check-e-p (λ () (read-dict-file "library/a.dict" x))
                          'read-dict-file))))

(define-test-suite repl-suite
  (current-procedures (list base-procedures))
  (test-case "repl"
             (check-equal? (with-input-from-string "${+ 1\n2}"
                             (λ () (with-output-to-string repl)))
                           "\n> | 3\n> ")
             (check-equal? (with-input-from-string "${concat '\n'}"
                             (λ () (with-output-to-string repl)))
                           "\n> | \n\n> ")
             (check-equal? (with-input-from-string "${concat \\\n\\\n}"
                             (λ () (with-output-to-string repl)))
                           "\n> | | \n\n\n> ")))

(define (remove-time str)
  (regexp-replace* #px"[0-9]{2}:[0-9]{2}:[0-9]{2} " str ""))

(define number-of-generators 13)

(define-test-suite system-suite
  (when (directory-exists? (output-dir))
    (delete-directory/files (output-dir)))
  (current-procedures (list base-procedures system-procedures))
  (current-cache (make-cache))
  (test-case "for-each-file"
             (let* ([p (lambda (dir max-depth)
                         (let ([x '()])
                           (for-each-file (lambda (y)
                                            (when (dict-file? y)
                                              (set! x (cons (path->string y)
                                                            x))))
                                          dir #:max-depth max-depth)
                           (map string->path (sort x string<?))))]
                    [x (p "library" 0)])
               (check-equal? x (map normal-case-path
                                    '("a.dict" "b.dict" "c/d.dict")))
               (check-equal? (p "library" 2) x)
               (check-equal? (p "library" 1)
                             (map normal-case-path '("a.dict" "b.dict")))))
  (test-case "read-dict-directory"
             (let ([x (read-dict-directory "library")])
               (check-equal? (hash-count x) 5)
               (check-equal? ((hash-ref x 'e)) "5")
               (check-equal? ((hash-ref x 'd)) "4")))
  (test-case "update-cache/generate? 1"
             (check-equal? (hash-count (cache-generators (current-cache)))
                           0)
             (check-true (update-cache/generate?))
             (check-equal? (hash-count (cache-generators (current-cache)))
                           number-of-generators))
  (test-case "library-ref"
             (check-equal? (library-ref) (cache-library (current-cache)))
             (check-equal? ((hash-ref (library-ref) 'e)) "5"))
  (test-case "generator-ref"
             (check-e-p (λ () (generator-ref 'x)) 'unknown-generator)
             (check-equal? (parameterize ([current-procedures
                                           (list* (generator-ref 'a)
                                                  (library-ref)
                                                  (current-procedures))])
                             (procedure-ref/call 'output))
                           "4"))
  (test-case "dict-path->symbol"
             (let ([d-p->s dict-path->symbol]
                   [d-p->s* (curryr dict-path->symbol #:remove-suffix? #f)]
                   [p->s (compose string->symbol path->string)])
               (check-equal? (d-p->s "a.dict") 'a)
               (check-equal? (d-p->s* "a.dict") 'a.dict)
               (check-equal? (d-p->s* "") '||)
               (check-equal? (d-p->s* ".") '||)
               (check-equal? (d-p->s* "/") '||)
               (check-equal? (d-p->s* "./") '||)
               (check-equal? (d-p->s* "/./") '||)
               (check-equal? (d-p->s "/a//b.dict/")
                             (p->s (build-path "a" "b")))
               (check-equal? (d-p->s* "//a/b/c/d.dict///")
                             (p->s (if (eq? (system-type) 'windows)
                                       (build-path "c" "d.dict")
                                       (build-path "a" "b" "c" "d.dict"))))))
  (test-case "safe-path"
             (let ([s-p safe-path])
               (check-equal? (s-p "a/b") "a/b")
               (check-e-p (λ () (s-p "..")) 'directory-traversal)
               (check-e-p (λ () (s-p "/..")) 'directory-traversal)
               (check-e-p (λ () (s-p "/../")) 'directory-traversal)
               (check-e-p (λ () (s-p "/../")) 'directory-traversal)
               (check-e-p (λ () (s-p "b/..")) 'directory-traversal)
               (check-e-p (λ () (s-p "b/../")) 'directory-traversal)
               (check-e-p (λ () (s-p "b/../c")) 'directory-traversal)
               (check-e-p (λ () (s-p "/..")) 'directory-traversal)))
  (test-case "extern"
             (check-e-p (λ () (T "${a | extern x}")) 'unknown-generator)
             (check-e-p (λ () (T "${a | extern ..}")) 'directory-traversal)
             (check-equal? (T "${a | extern b/c}") "2")
             (check-equal? (T "${a | extern b/c}") "2")
             (check-equal? (T "${a | extern /b/c}") "2")
             (check-equal? (T "${a | extern ./b/c}") "2")
             (check-equal? (T "${a | extern b/c/}") "2")
             (check-equal? (T "${a | extern b//c/}") "2")
             (check-equal? (T "${a | extern a}") "-1")
             (check-equal? (hash-count (cache-generators (current-cache)))
                           number-of-generators))
  (test-case "generator"
             (check-e-p (λ () (T "${generator}")) 'unknown)
             (check-equal? (T "${generator | extern b/c}")
                           (path->string (build-path "b" "c"))))
  (test-case "generators"
             (check-e-p (λ () (T "${generators k}")) 'unexpected-sequence)
             (check-e-p (λ () (T "${generators k ..}")) 'directory-traversal)
             (check-e-p (λ () (T "${generators k '' x}")) 'invalid-integer)
             (check-equal? (T "${generators k | length}")
                           (number->string (- number-of-generators 3)))
             (check-equal? (T "${generators k b | length}") "2")
             (check-equal? (T "${generators k ./b | length}") "2")
             (check-equal? (T "${generators k /b | length}") "2")
             (check-equal? (T "${generators k //b// | length}") "2")
             (check-equal? (T "${generators k b | sort {a | k}"
                              " | join {a | k}}")
                           "23")
             (check-equal? (T "${generators k b 0 | sort {a | k}"
                              " | join {a | k}}")
                           "234")
             (check-equal? (T "${generators k b |sort {generator | k}"
                              " | join {generator | k} ,}")
                           (path->string (build-path "b" "c,b" "d")))
             (check-equal? (hash-count (cache-generators (current-cache)))
                           number-of-generators))
  (test-false "update-cache/generate? 2" (update-cache/generate?))
  (test-case "copy-static-files"
             (parameterize ([dummy? #t]) (copy-static-files))
             (let* ([static2 (build-path "b" "static2")]
                    [static2* (build-path (output-dir) static2)]
                    [static1 (build-path (output-dir) "static1")])
               (check-false (file-exists? static2*))
               (parameterize ([omit-file? (curry regexp-match #rx"[^2]$")])
                 (check-equal? (remove-time (with-output-to-string
                                              (λ () (copy-static-files)
                                                 (copy-static-files))))
                               (format "~a~%" static2)))
               (check-true (file-exists? static2*))
               (check-false (file-exists? static1))
               (parameterize ([quiet? #t])
                 (check-equal? (with-output-to-string copy-static-files) ""))
               (check-true (file-exists? static1))))
  (test-case "generate-variants path"
             (let* ([p "path.txt"]
                    [q (build-path (output-dir) p)]
                    [g (lambda ()
                         (with-output-to-string
                           (λ () (generate-variants 'path.txt))))])
               (parameterize ([dummy? #t])
                 (check-equal? (remove-time (g)) (format "~a~%" p)))
               (check-false (file-exists? q))
               (parameterize ([quiet? #t])
                 (check-equal? (g) ""))
               (check-true (file-exists? q))))
  (test-case "generate-variants index-file"
             (let ([p (build-path "x" "INDEX-FILE")])
               (check-equal? (remove-time
                              (with-output-to-string
                                (λ () (generate-variants 'index-file))))
                             (format "index-file -> ~a~%" p))
               (check-true (file-exists? (build-path (output-dir) p)))))
  (test-case "generate-variants traversal1"
             (check-e-p (λ () (generate-variants 'traversal1))
                        '(generate-variant directory-traversal)))
  (test-case "generate-variants traversal2"
             (check-e-p (λ () (generate-variants 'traversal2))
                        '(generate-variant directory-traversal)))
  (test-case
   "generate-variants input"
   (let* ([files '("xa" "xb" "xc")]
          [files* (map (curry build-path (output-dir)) files)]
          [p (lambda (fs)
               (apply string-append (map (curry format "input -> ~a~%")
                                         fs)))])
     (check-equal? (remove-time (with-output-to-string
                                  (λ () (generate-variants 'input))))
                   (p files))
     (for ([file files*]) (check-true (file-exists? file)))
     (check-equal? (remove-time
                    (with-output-to-string
                      (lambda ()
                        (generate-variants 'input (string->thunk
                                                   "${input | reverse}")))))
                   (p (reverse files)))))
  (quiet? #t)
  (test-case "generate-variants bytes"
             (generate-variants 'bytes)
             (check-equal? (file-size (build-path (output-dir) "bytes")) 1)
             (check-e-p
              (lambda ()
                (generate-variants
                 'bytes (string->thunk
                         "${yield | extended output {char 256}}")))
              (list 'generate-variant 'bytes 'file
                    (build-path "output" "bytes") 'encoding-error
                    (string (integer->char 256)))))
  (test-case "generate-variants utf8"
             (generate-variants 'utf8)
             (check-equal? (file-size (build-path (output-dir) "utf8")) 2))
  (test-case "generate-variants mode"
             (generate-variants 'mode)
             (check-equal? (file-or-directory-permissions
                            (build-path (output-dir) "mode"))
                           '(execute write read)))
  (test-case "generate-variants write"
             (check-e-p (λ () (procedure-ref/call 'write (λ () "")))
                        'not-generating)
             (parameterize ([dummy? #t])
               (generate-variants 'write))
             (let ([f (build-path (output-dir) "write")])
               (check-false (file-exists? f))
               (generate-variants 'write)
               (check-equal? (call-with-input-file f (curry read-string 4))
                             "abcd")))
  (test-case
   "input-files"
   (check-equal? (string->number (T "${input-files f '' 0"
                                    "| filter {f | pregexp-match? [.]dict$}"
                                    "| length}"))
                 number-of-generators)
   (check-true (< (string->number (T "${input-files f b"
                                     "| filter {f | pregexp-match? [.]dict$}"
                                     "| length}"))
                  (string->number (T "${input-files f"
                                     "| filter {f | pregexp-match? [.]dict$}"
                                     "| length}"))
                  number-of-generators))
   (check-e-p (λ () (T "${input-files x}")) 'unexpected-sequence)
   (check-e-p (λ () (T "${input-files x ..}")) 'directory-traversal)
   (check-e-p (λ () (T "${input-files x '' x}")) 'invalid-integer))
  (test-case "input-file-size"
             (check-equal? (T "${input-file-size utf8.dict}")
                           "10")
             (check-e-p (λ () (T "${input-file-size ..}"))
                        'directory-traversal)
             (check-e-p (λ () (T "${input-file-size x}"))
                        'unknown-input-file))
  (test-case "input-file-string"
             (check-equal? (T "${input-file-string utf8.dict}")
                           (file->string "input/utf8.dict"))
             (check-e-p (λ () (T "${input-file-string ..}"))
                        'directory-traversal)
             (check-e-p (λ () (T "${input-file-string x}"))
                        'unknown-input-file))
  (test-case "input-file-bytes"
             (check-equal? (T "${input-file-bytes utf8.dict}")
                           (bytes->string/latin-1
                            (file->bytes "input/utf8.dict")))
             (check-e-p (λ () (T "${input-file-bytes ..}"))
                        'directory-traversal)
             (check-e-p (λ () (T "${input-file-bytes x}"))
                        'unknown-input-file))
  (test-case "input-file-timestamp"
             (check-e-p (λ () (T "${input-file-timestamp ..}"))
                        'directory-traversal)
             (check-e-p (λ () (T "${input-file-timestamp x}"))
                        'unknown-input-file)
             (check-true (>= (string-length
                              (T "${input-file-timestamp utf8.dict true}"))
                             (string-length
                              (T "${input-file-timestamp utf8.dict}"))
                             20))))

(define-test-suite all-tests
  runtime-suite
  base-suite
  procedure-suite
  dict-suite
  repl-suite
  system-suite)

(let ([temp (build-path (find-system-path 'temp-dir)
                        (number->string (current-seconds)))]
      [default-error-display-handler (error-display-handler)])
  (make-directory temp)
  (parameterize ([error-display-handler
                  (lambda (msg exn)
                    (default-error-display-handler msg exn)
                    (when (error/params? exn)
                      (print-error/params exn)))]
                 [current-directory temp])
    (for ([(path content)
           (hash "input/a.dict" "|a -1\n|output ${+ {a} {b} {c}}"
                 "input/b/c.dict" "|a 2\n"
                 "input/b/d.dict" "|a 3"
                 "input/b/e/f.dict" "|a 4"
                 "input/b/static2" ""
                 "input/bytes.dict" "|bytes? true\n|output ÿ"
                 "input/index-file.dict"
                 "|path x/\n|index-file ${generator | upcase}\n|output"
                 "input/input.dict"
                 "|input ${args x a b c}\n|path x${x}\n|output"
                 "input/mode.dict" "|mode #o777\n|output"
                 "input/path.txt.dict" "|output ${a}"
                 "input/static1" ""
                 "input/traversal1.dict"
                 "|path x/\n|index-file ../../y\n|output"
                 "input/traversal2.dict" "|path ../x\n|output"
                 "input/utf8.dict" "|output ÿ"
                 "input/write.dict" "|output\nc${write a}${write b}d"
                 "library/a.dict"
                 "|# foo\n|# bar\n|a 1\n|#\nfoo\nbar\n|b 2\n|c 3\n|#"
                 "library/b.dict" "|d 4"
                 "library/c/d.dict" "|e 5")])
      (make-directory* (path-only path))
      (call-with-output-file path (curry display content)))
    (run-tests all-tests))
  (delete-directory/files temp))
